/*
 * Copyright (C) 2009 Voltaire Ltd. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#define _GNU_SOURCE
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <complib/cl_types.h>
#include <complib/cl_debug.h>
#include <infiniband/umad.h>
#include <infiniband/mad.h>
#include <signal.h>
#include <syslog.h>
#include <http_client.h>

#define PACK_SUFFIX __attribute__((packed))
#define IB_MLX_CONFIG_SPACE_ATTR_MOD ((2 << 22) | (12 << 16)) /* 12 record-ports at time */
#define IB_MLX_VENDOR_CLASS	10
#define IB_MLX_IS3_GENERAL_INFO		0x17
#define IB_MLX_IS3_CONFIG_SPACE_ACCESS	0x50
#define NUM_HW_DEV			28
#define NUM_SW_PORTS		25 /* port 0 included */
#define MAX_SIZE		20000
#define MAX_SW_SIZE		5000
#define WINDOW_SIZE		32
#define D_LEASETIME		300
#define UMAD_SIZE		512
#define D_ADDR			"127.0.0.1"
#define D_PORT			8005
#define D_KEY			0x12345
#define DEFAULT_SAMPLE_INTERVAL 5
#define DEFAULT_THRESHOLD 80
#define DEFAULT_CONF_FILENAME	"/etc/ibpm/ibpm.conf"
#define DEFAULT_STATS_FILENAME	"/var/log/ibpm/ibpm.stats"
#define DEFAULT_REC_FILENAME    "/etc/ibpm/records.conf"
#define NO_EXTENDED_PC_FILENAME	"/etc/ibpm/epc_devid.deny"
#define DEFAULT_LOG_FILENAME    "/var/log/ibpm/ibpm.log"
#define DEFAULT_CSV_FILENAME    "/var/log/ibpm/csv/"
#define DEFAULT_TIMEOUT       50
/* Default poll interval for port state */
#define DEFAULT_POLL_INTERVAL 5
/* Default poll retries for port state */
#define DEFAULT_POLL_RETRIES 12
#define IBPM_VERSION		"1.1.3"
//#define NUM_CLASSES		3
#define MAX_CLASS		256
#define MSG_DATA_LEN	65507 /* max message len is the max UDP message size */
#define MSG_DATA_LEN_THR (MSG_DATA_LEN-500)
#define _GNU_SOURCE		/* for strsignal */

enum {
	IBPM_ERR,
	IBPM_WARN,
	IBPM_INFO,
	IBPM_DEBUG
};

enum {
	MAD_TYPE_PC,
	MAD_TYPE_PC_EXT,
	MAD_TYPE_LOW_XMITWAIT,
	MAD_TYPE_HI_XMITWAIT,
	MAD_TYPE_MAX
};

enum {
	RESET_EVERY_POLL,
	RESET_ON_THRESHOLD,
	NO_RESET
};


char *argv0 = "ibpm";

struct perf_count {
	uint32_t portselect;
	uint32_t counterselect;
	uint32_t symbolerrors;
	uint32_t linkrecovers;
	uint32_t linkdowned;
	uint32_t rcverrors;
	uint32_t rcvremotephyerrors;
	uint32_t rcvswrelayerrors;
	uint32_t xmtdiscards;
	uint32_t xmtconstrainterrors;
	uint32_t rcvconstrainterrors;
	uint32_t linkintegrityerrors;
	uint32_t excbufoverrunerrors;
	uint32_t vl15dropped;
	uint64_t xmtwait;
	uint64_t xmtdata;
	uint64_t rcvdata;
	uint64_t xmtpkts;
	uint64_t rcvpkts;
};

struct perf_count_ext {
	uint32_t portselect;
	uint32_t counterselect;
	uint64_t portxmitdata;
	uint64_t portrcvdata;
	uint64_t portxmitpkts;
	uint64_t portrcvpkts;
	uint64_t portunicastxmitpkts;
	uint64_t portunicastrcvpkts;
	uint64_t portmulticastxmitpkits;
	uint64_t portmulticastrcvpkts;
};

typedef struct {
	uint32_t address;
	uint32_t data;
	uint32_t mask;
} is3_record_t;

typedef struct {
	uint16_t hw_revision;
	uint16_t device_id;
	uint8_t  reserved[24];
	uint32_t uptime;
} is3_hw_info_t;

typedef struct {
	uint8_t resv1;
	uint8_t major;
	uint8_t minor;
	uint8_t sub_minor;
	uint32_t build_id;
	uint8_t month;
	uint8_t day;
	uint16_t year;
	uint16_t resv2;
	uint16_t hour;
	uint8_t psid[16];
	uint32_t ini_file_version;
} is3_fw_info_t;

typedef struct {
	uint8_t resv1;
	uint8_t major;
	uint8_t minor;
	uint8_t sub_minor;
	uint8_t resv2[28];
} is3_sw_info_t;

typedef struct {
	uint8_t reserved[8];
	is3_hw_info_t hw_info;
	is3_fw_info_t fw_info;
	is3_sw_info_t sw_info;
} is3_general_info_t;

typedef struct {
	uint8_t reserved[8];
	is3_record_t record[18];
} is3_config_space_t;

struct perf_count perf_count = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
struct perf_count_ext perf_count_ext = {0,0,0,0,0,0,0,0,0,0};

typedef struct sw_rec {
	unsigned int lid;
	unsigned int devid;
	unsigned int rev;
	uint64_t xmitwait[25];
	uint64_t xmitwait_cur_sum[25];
	uint64_t xmitwait_new_sum[25];
} sw_rec_t;

typedef struct record {
	/* rec type
	0. pc
	1. pc_ext
	2. xmit_wait low counters
	3. xmit_wait high counters
	*/
	unsigned int type;
	/* record's sw_id points to vendor record in rarray
	and vendor record's sw_id points to switch record in sarray
	*/
	unsigned int sw_id;
	unsigned int index;
	unsigned int lid;
	unsigned int port;
	unsigned int ireport;
	unsigned int isample;
	unsigned int devid;
	unsigned int rev;
	unsigned int vendor_xmitwait;
	uint64_t guid;
	unsigned int to_report;
	unsigned int firstrun;
	int dt;
	unsigned int status;
	uint32_t reset_trid[MAD_TYPE_MAX];
	unsigned int pc_timeouts;	  /* port counters MADs timeouts */
	unsigned int socket_timeouts; /* socket timeouts */
	struct timeval cur_time;
	struct timeval send_time;
	struct timeval last_time;
	struct perf_count pcount;
	struct perf_count cur_pc_sum;
	struct perf_count new_pc_sum;
	boolean_t is_support_extended_pc;
	unsigned int ext_id;
} record_t;

typedef struct log {
	int   dbg_lvl;
	FILE *file;
	char *filename;
	FILE *csvfile;
} log_t;

typedef struct config {
	FILE *file;
	char *filename;
	time_t mtime;
} config_t;

typedef struct vlr_config {
	char *psl_file;
	int  psl_cmdln; /*if 1 do not read psl_file from conf file */
	time_t psl_mtime;
	uint8_t  *lid_to_sl;
	int psl_rc;

	char *peer_path_file;
	int  peer_cmdln; /*if 1 do not read peer_path_file from conf file */
	time_t peer_path_mtime;
	uint16_t *ca_to_sw_lid;
	int peer_rc;
} vlr_config_t;

/*
Header
class		(32 bits)
version	(8  bits)
type		(8  bits)
ufm_key	(64 bits)
source_guid (64 bits)
source_port (8  bits)
time_sec	(64 bits)
time_usec	(64 bits)
*/
typedef struct header {
	char msg_class[4];
	char type;
	char version;
	uint64_t ufm_key;
	uint64_t source_guid;
	char source_port;
	uint64_t sec;
	uint64_t usec;
} PACK_SUFFIX header_t;

typedef struct message {
	header_t header;
	char data[MSG_DATA_LEN];
} PACK_SUFFIX message_t;

typedef struct {
	uint32_t trid;
	struct timeval time;
	int index;
}trid_arr_t;

typedef struct tables {
	record_t **rlinks;
	record_t *rarray;
	sw_rec_t *sarray;
	int rsize;
	int ssize;
	int numrecords;
} tables_t;

typedef struct stats {
	unsigned int pc_timeouts;	  /* port counters MADs timeouts */
	unsigned int vendor_timeouts; /* vendor specifc MADs timeouts */
	unsigned int socket_timeouts; /* socket timeouts */
	unsigned int mad_sent;		  /* total sent MADs */
	unsigned int avg_send_time;   /* avg MAD send time (usec) */
	unsigned int max_avg;		  /* maximum avg time */
	unsigned int min_avg;		  /* minimum avg time */
	unsigned int switch_tbl;	  /* amount of table switches */
	unsigned int frag_msg;		  /* each time msg is fragged */
	unsigned int table_diff;	  /* diff between last two tables */
	unsigned int reset_error_pc;    /* amount of reset error counters */
	unsigned int dropped;
} stats_t;

typedef struct extended_pc_info {
	unsigned int devid;	     /* deviceid */
	boolean_t epc_supported; /* true if epc is suported */
} extended_pc_info_t;

#define MAX_DEVID_NUMBER 128
#define UFM_PROTOCOL_LEN 16
#define UFM_URL_LEN 64
#define UFM_REST_URL_LEN 256

typedef struct ibpm {
	stats_t stats;
	log_t log;
	config_t config;
	int active_tbl;
	tables_t table[2];
	message_t *message;
	char *ca;
	int update_socket;
	int window_size;
	int report_level;
	int read_conf;
	int print_stats;
	int reset_stats;
	int ca_port;
	int sent;
	int csv;
	int timeout;
	int finished;
	int numrecords;
	int eraselog;
	int verbose;
	int leasetime;
	int default_poll_time;
	int default_poll_time_msec;
	int reset_mode;
	int threshold;
	int errrate;
	int port_poll_interval;
	int port_poll_retries;
	int clear_all[2];
	char address[16];
	char ufm_protocol[UFM_PROTOCOL_LEN];
	char ufm_url[UFM_URL_LEN];
	char ufm_rest_url[UFM_REST_URL_LEN];
	uint8_t umad[UMAD_SIZE];
	char data[1024];
	int port;
	int sockfd;
	struct sockaddr_in servaddr;
	HTTP_HANDLE httphandle;
	trid_arr_t *trid_arr;
	struct ibmad_port *srcport;
	struct perf_count pc_thr;
	extended_pc_info_t epc_support_info[MAX_DEVID_NUMBER];
	int epc_support_info_size;
	unsigned int max_timeouts_number;
	uint16_t baselid;
	vlr_config_t vlr_config;
} ibpm_t;


/*
#ifdef __DEBUG
	#define IBPM_DEBUG(log, level, fmt, arg...) do { \
		if (ibmp_log_is_active(log, (level))) \
			ibpm_log(log, level, "%s: " fmt, __func__, ##arg); \
		} while (0)
#else
	#define IBPM_DEBUG(log, level, fmt, arg...)
#endif
*/

#define IBPM_LOG(log, level, fmt, arg...) do { \
	if (ibpm_dbg_active(log, (level))) \
		ibpm_log(log, level, "%s: " fmt, __func__, ##arg); \
	} while (0)
