/*
 * Copyright (C) 2009 Voltaire Ltd. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 *
 *	Author: Samer Deeb
 *	Date: Mar, 3 2012
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <curl/curl.h>
#include <curl/easy.h>

#include "http_client.h"

static int g_GlobalInit = 0;
static char * OPEN_SSL_CIPHER = "RC4-SHA";
static char * NSS_CIPHER = "rsa_rc4_128_sha";

typedef struct _MemoryHandle {
  char *	memory;
  size_t 	size;
}MemoryHandle;

typedef struct _HTTPHandle_t {
	MemoryHandle * 	writeMem;
	MemoryHandle * 	readMem;
	CURL * 			curlPtr;
}HTTPHandle;

static void *memrealloc(void *ptr, size_t size)
{
  /* There might be a realloc() out there that doesn't like reallocing
     NULL pointers, so we take care of it here */
  if(ptr)
    return realloc(ptr, size);
  else
    return malloc(size);
}

static void memfree(MemoryHandle * mem)
{
	if(mem)
	{
		if (mem->memory)
			free(mem->memory);
		free(mem);
	}
}

static void memempty(MemoryHandle * mem)
{
	if(mem)
	{
		mem->memory = NULL;
		mem->size = 0;
	}
}

static size_t
WriteMemoryCallback(void *ptr, size_t size, size_t nmemb, void *data)
{
	size_t realsize = size * nmemb;
	MemoryHandle *mem = (MemoryHandle *)data;

	mem->memory = memrealloc(mem->memory, mem->size + realsize + 1);
	if (mem->memory)
	{
		memcpy(&(mem->memory[mem->size]), ptr, realsize);
		mem->size += realsize;
		mem->memory[mem->size] = 0;
	}
	return realsize;
}

static size_t ReadCallback(void *ptr, size_t size, size_t nmemb, void *stream)
{
	size_t bytessent = 0;
	size_t totalsize = size * nmemb;
	MemoryHandle * mem = (MemoryHandle *) stream;

	/* in real-world cases, this would probably get this data differently
       as this fread() stuff is exactly what the library already would do
       by default internally */
	printf("ReadCallback: %d", (int)mem->size);
	if (mem->size == 0)
	{
		*((char *)ptr) = '\0';
		return 0;
	}

	bytessent = totalsize > mem->size ?mem->size: totalsize;

	memcpy((char *)ptr, mem->memory, bytessent);
	mem->memory += bytessent;
	mem->size -= bytessent;
	return bytessent;
}

static char * getCipher()
{
  if(strstr(curl_version(), " NSS/") != NULL)
	  return NSS_CIPHER;
  return OPEN_SSL_CIPHER;
}

static void resetCurl(CURL * curl)
{
	curl_easy_setopt(curl, CURLOPT_NOBODY, 1L);
	curl_easy_setopt(curl, CURLOPT_INFILESIZE, 0);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, 0);
	curl_easy_setopt(curl, CURLOPT_POSTFIELDS, NULL);
}

HTTP_HANDLE HTTPClient_Init(int use_ssl)
{
	HTTPHandle *curl_handle;

	if (!g_GlobalInit)
	{
		curl_global_init(CURL_GLOBAL_ALL);
		g_GlobalInit = 1;
	}

	/* init the http session */
	curl_handle = malloc(sizeof(HTTPHandle));
	/*init memory handles for read/write */
	curl_handle->writeMem = calloc(1, sizeof(MemoryHandle));
	curl_handle->readMem = calloc(1, sizeof(MemoryHandle));
	/* init the curl session */
	curl_handle->curlPtr = curl_easy_init();
	if (use_ssl)
	{
		/* init ssl */
		curl_easy_setopt(curl_handle->curlPtr, CURLOPT_SSL_VERIFYPEER, 0L);
		curl_easy_setopt(curl_handle->curlPtr, CURLOPT_SSL_VERIFYHOST, 0L);
		curl_easy_setopt(curl_handle->curlPtr, CURLOPT_SSL_CIPHER_LIST, getCipher());
	}
	curl_easy_setopt(curl_handle->curlPtr, CURLOPT_VERBOSE, 1L);

	/* write all data to this function  */
    curl_easy_setopt(curl_handle->curlPtr, CURLOPT_WRITEFUNCTION, WriteMemoryCallback);
    /* set write data handler */
    curl_easy_setopt(curl_handle->curlPtr, CURLOPT_WRITEDATA, curl_handle->writeMem);

    /* read all data from this function  */
	curl_easy_setopt(curl_handle->curlPtr, CURLOPT_READFUNCTION, ReadCallback);
	/* set read data handler */
	curl_easy_setopt(curl_handle->curlPtr, CURLOPT_READDATA, curl_handle->readMem);


	return (HTTP_HANDLE) curl_handle;
}

int HTTPClient_setAuth(HTTP_HANDLE handle, const char * user, const char * pwd)
{
	CURL * curl = NULL;
	char * user_pwd = NULL;
	if (handle)
		curl = ((HTTPHandle *)handle)->curlPtr;
	if(!curl)
		return HTTPClient_NotInited;
	if (!user || !pwd)
		return HTTPClient_InvalidArg;

	user_pwd = (char *)malloc(strlen(user) + strlen(pwd) + 2);
	sprintf(user_pwd, "%s:%s", user, pwd);

	curl_easy_setopt(curl, CURLOPT_USERPWD, user_pwd);
	return HTTPClient_OK;
}

int HTTPClient_get(HTTP_HANDLE handle, const char * url, char **data, int * plen, long * presponse)
{
	CURL * curl = NULL;
	int res;

	if (handle)
		curl = ((HTTPHandle *)handle)->curlPtr;
	if(!curl)
		return HTTPClient_NotInited;

	if (!data || ! plen || ! presponse)
		return HTTPClient_InvalidArg;

	*presponse = 0;
	*data = NULL;
	*plen = 0;

	/* set HTTP Method - GET*/
	curl_easy_setopt(curl, CURLOPT_HTTPGET, 1);
	curl_easy_setopt(curl, CURLOPT_URL, url);
	res = curl_easy_perform(curl);
	if (res == 0)
	{
		/* get response code */
		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, presponse);
		/* get data */
		*data =  ((HTTPHandle *)handle)->writeMem->memory;
		*plen =  ((HTTPHandle *)handle)->writeMem->size;
		return HTTPClient_OK;
	}

	return HTTPClient_Fail;
}

int HTTPClient_put(HTTP_HANDLE handle, const char * url, char *data, int len, long * presponse)
{
	CURL * curl = NULL;
	int res;
	MemoryHandle * readMem = NULL;

	if (handle)
		curl = ((HTTPHandle *)handle)->curlPtr;
	if(!curl)
		return HTTPClient_NotInited;

	if (data && len)
	{
		/* enable body */
		curl_easy_setopt(curl, CURLOPT_NOBODY, 0L);
		/*set file size*/
		curl_easy_setopt(curl, CURLOPT_INFILESIZE, len);

		/*set body data*/
		readMem = ((HTTPHandle *)handle)->readMem;
		readMem->memory = data;
		readMem->size = len;
	}
	else
	{
		curl_easy_setopt(curl, CURLOPT_INFILESIZE, 0);
	}

	/* set HTTP Method - PUT*/
	curl_easy_setopt(curl, CURLOPT_PUT, 1L);
	curl_easy_setopt(curl, CURLOPT_URL, url);
	res = curl_easy_perform(curl);
	memempty(readMem);
	resetCurl(curl);
	if (res == 0)
	{
		/* get response code */
		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, presponse);
		return HTTPClient_OK;
	}

	return HTTPClient_Fail;
}

int HTTPClient_post(HTTP_HANDLE handle, const char * url, char *data, int len, long * presponse)
{
	CURL * curl = NULL;
	MemoryHandle * readMem = NULL;
	int res;

	if (handle)
		curl = ((HTTPHandle *)handle)->curlPtr;
	if(!curl)
		return HTTPClient_NotInited;

	if (data && len)
	{
		/* enable body */
		curl_easy_setopt(curl, CURLOPT_NOBODY, 0L);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, (curl_off_t)len);
		curl_easy_setopt(curl, CURLOPT_INFILESIZE, len);
		/*set body data*/
		readMem = ((HTTPHandle *)handle)->readMem;
		readMem->memory = data;
		readMem->size = len;
	}
	else
	{
		curl_easy_setopt(curl, CURLOPT_NOBODY, 1L);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDSIZE, 0);
		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, NULL);
	}

	/* set HTTP Method - POST*/
	curl_easy_setopt(curl, CURLOPT_POST, 1L);
	curl_easy_setopt(curl, CURLOPT_URL, url);
	res = curl_easy_perform(curl);

	resetCurl(curl);

	memempty(readMem);
	if (res == 0)
	{
		/* get response code */
		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, presponse);
		return HTTPClient_OK;
	}

	return HTTPClient_Fail;
}

int HTTPClient_delete(HTTP_HANDLE handle, const char * url, long * presponse)
{
	CURL * curl = NULL;
	int res;

	if (handle)
		curl = ((HTTPHandle *)handle)->curlPtr;
	if(!curl)
		return HTTPClient_NotInited;

	/* disable body */
	curl_easy_setopt(curl, CURLOPT_NOBODY, 1L);

	/* set HTTP Method - DELETE*/
	curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, "DELETE");
	curl_easy_setopt(curl, CURLOPT_URL, url);
	res = curl_easy_perform(curl);
	curl_easy_setopt(curl, CURLOPT_CUSTOMREQUEST, NULL);
	if (res == 0)
	{
		/* get response code */
		curl_easy_getinfo(curl, CURLINFO_RESPONSE_CODE, presponse);
		return HTTPClient_OK;
	}

	return HTTPClient_Fail;
}

void HTTPClient_cleanup(HTTP_HANDLE handle)
{
	CURL * curl = NULL;

	if (!handle)
		return;

	curl = ((HTTPHandle *)handle)->curlPtr;
	if(curl)
		 curl_easy_cleanup(curl);

	memfree(((HTTPHandle *)handle)->readMem);
	memfree(((HTTPHandle *)handle)->writeMem);
	free(handle);
}
